/****************************************************************************
 * Copyright (c) 2010-2012, EBM WebSourcing - All rights reserved.
 * Copyright (c) 2013, Linagora - All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/
 
package com.ebmwebsourcing.easycommons.lang.reflect;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.junit.Test;

import com.ebmwebsourcing.easycommons.lang.UncheckedException;

/**
 * @author Adrien Ruffie - EBM WebSourcing
 */
public class ReflectionHelperTest {

    private static interface InterfaceA {
    }

    private static interface InterfaceB {
        void methodA();
    }

    private static interface InterfaceC {
    }

    private static interface InterfaceD {
    }

    private static interface InterfaceE extends InterfaceC {
    }

    private static abstract class ClassA implements InterfaceA, InterfaceB {
        public ClassA() {
        }
    }

    private static class ClassB extends ClassA {

        private ClassB() {
        }

        @Override
        public final void methodA() {
        }

        @SuppressWarnings("unused")
        private final void methodB(int param1) {
        }

        @SuppressWarnings("unused")
        public final void methodC(int param1) {
        }

    }

    private static final class ClassC extends ClassB implements InterfaceA, InterfaceE, InterfaceD {
    }

    public class ClassE {

        @SuppressWarnings("unused")
        private void methodA(Integer param1) {
        }

        @SuppressWarnings("unused")
        private void methodList(final List<String> param1) {
        }

        public String methodString() {
            return "0";
        }

        public int methodInt() {
            return 0;
        }
    }

    @Test
    public void testFindAllImplementedInterfaces0() throws Exception {
        Set<Class<?>> implementedInterfaces = ReflectionHelper
                .findAllImplementedInterfaces(ClassD.class);
        Set<Class<?>> expectedImplementedInterfaces = new HashSet<Class<?>>();
        assertEquals(expectedImplementedInterfaces, implementedInterfaces);
    }

    @Test
    public void testFindAllImplementedInterfaces1() throws Exception {
        Set<Class<?>> implementedInterfaces = ReflectionHelper
                .findAllImplementedInterfaces(ClassA.class);
        Set<Class<?>> expectedImplementedInterfaces = new HashSet<Class<?>>();
        expectedImplementedInterfaces.add(InterfaceA.class);
        expectedImplementedInterfaces.add(InterfaceB.class);
        assertEquals(expectedImplementedInterfaces, implementedInterfaces);
    }

    @Test
    public void testFindAllImplementedInterfaces2() throws Exception {
        Set<Class<?>> implementedInterfaces = ReflectionHelper
                .findAllImplementedInterfaces(ClassB.class);
        Set<Class<?>> expectedImplementedInterfaces = new HashSet<Class<?>>();
        expectedImplementedInterfaces.add(InterfaceA.class);
        expectedImplementedInterfaces.add(InterfaceB.class);
        assertEquals(expectedImplementedInterfaces, implementedInterfaces);
    }

    @Test
    public void testFindAllImplementedInterfaces3() throws Exception {
        Set<Class<?>> implementedInterfaces = ReflectionHelper
                .findAllImplementedInterfaces(ClassC.class);
        Set<Class<?>> expectedImplementedInterfaces = new HashSet<Class<?>>();
        expectedImplementedInterfaces.add(InterfaceA.class);
        expectedImplementedInterfaces.add(InterfaceB.class);
        expectedImplementedInterfaces.add(InterfaceC.class);
        expectedImplementedInterfaces.add(InterfaceD.class);
        expectedImplementedInterfaces.add(InterfaceE.class);
        assertEquals(expectedImplementedInterfaces, implementedInterfaces);
    }

    @Test
    public void testFindMethodThatReturnString() throws Exception {
        Collection<Method> methods = ReflectionHelper.findMethodsThatReturnType(ClassE.class,
                String.class);

        assertEquals(2, methods.size());
        Iterator<Method> it = methods.iterator();
        assertEquals("methodString", it.next().getName());
        assertEquals("toString", it.next().getName());
    }

    @Test
    public void testFindMethodThatReturnInt() throws Exception {
        Collection<Method> methods = ReflectionHelper.findMethodsThatReturnType(ClassE.class,
                int.class);

        assertEquals(2, methods.size());
        Iterator<Method> it = methods.iterator();
        assertEquals("methodInt", it.next().getName());
        assertEquals("hashCode", it.next().getName());
    }

    @Test(expected = UncheckedException.class)
    public void testGetMethod0() throws Exception {
        ReflectionHelper.getPublicMethod(ClassB.class, "wrongMethodName");
    }

    @Test(expected = UncheckedException.class)
    public void testGetMethod1() throws Exception {
        ReflectionHelper.getPublicMethod(ClassB.class, "methodA", String.class);
    }

    @Test
    public void testGetMethod2() throws Exception {
        assertEquals(ClassB.class.getMethod("methodA"),
                ReflectionHelper.getPublicMethod(ClassB.class, "methodA"));
    }

    @Test(expected = UncheckedException.class)
    public void testGetMethod3() throws Exception {
        ReflectionHelper.getPublicMethod(ClassB.class, "methodB", int.class);
    }

    @Test
    public void testGetMethod4() throws Exception {
        ReflectionHelper.getPublicMethod(ClassB.class, "methodC", int.class);
    }

    @Test
    public void testGetPublicDeclaredMethods0() throws Exception {
        final List<Method> actualMethods = ReflectionHelper.getPublicDeclaredMethods(ClassE.class);

        assertEquals(2, actualMethods.size());
        final Iterator<Method> it = actualMethods.iterator();
        boolean methodStringFound = false;
        boolean methodIntFound = false;
        while (it.hasNext()) {
            final String methodName = it.next().getName();
            if ("methodString".equals(methodName)) {
                methodStringFound = true;
            } else if ("methodInt".equals(methodName)) {
                methodIntFound = true;
            } else {
                fail("Unexpected method name");
            }
        }
        assertTrue("methodString", methodStringFound);
        assertTrue("methodInt", methodIntFound);
    }

    @Test
    public void testGetPublicDeclaredMethods1() throws Exception {
        List<Method> actualMethods = ReflectionHelper.getPublicDeclaredMethods(ClassC.class);

        assertEquals(0, actualMethods.size());
    }

    @Test
    public void testinvokeMethod0() throws Exception {
        ClassB objectB = new ClassB();
        ReflectionHelper.invokeMethod(objectB,
                ReflectionHelper.getPublicMethod(ClassB.class, "methodA"));
    }

    @Test
    public void testinvokePrivateMethodDeprecated_0() throws Exception {
        ClassE objectE = new ClassE();
        ReflectionHelper.invokePrivateMethod(objectE, "methodA", 5);
    }

    @Test
    public void testinvokePrivateMethod0() throws Exception {
        ClassE objectE = new ClassE();
        ReflectionHelper.invokePrivateMethod(objectE, "methodA",
                new Object[] { Integer.valueOf(5) }, new Class[] { Integer.class });
    }

    @Test
    public void testinvokePrivateMethod1() throws Exception {
        ClassE objectE = new ClassE();
        final List<String> list = new ArrayList<String>();
        ReflectionHelper.invokePrivateMethod(objectE, "methodList", new Object[] { list },
                new Class[] { List.class });
    }

    @Test
    public void testinvokePrivateMethod_StaticPrivateMethod() throws Exception {
        final String string = "string";
        assertEquals(string, ReflectionHelper.invokePrivateMethod(ClassWithStaticMethods.class,
                null, "myMethod", new Object[] { string }, new Class[] { String.class }));
    }

    @Test(expected = UncheckedException.class)
    public void testinvokePrivateMethod_NullNull() throws Exception {
        ReflectionHelper.invokePrivateMethod(null, null, "myMethod", new Object[] {},
                new Class[] {});
    }

    @Test
    public void testinvokeMethod1() throws Exception {
        ClassB objectB = new ClassB();
        ReflectionHelper.invokeMethod(objectB,
                ReflectionHelper.getPublicMethod(ClassB.class, "methodC", int.class), 5);
    }

    @Test(expected = UncheckedException.class)
    public void testinvokeMethod2() throws Exception {
        ClassB objectB = new ClassB();
        ReflectionHelper.invokeMethod(objectB,
                ReflectionHelper.getPublicMethod(ClassB.class, "methodA"), "Illegal argument");
    }

    @Test(expected = UncheckedException.class)
    public void testinvokeMethod3() throws Exception {
        ClassB objectB = new ClassB();
        ReflectionHelper.invokeMethod(objectB,
                ClassB.class.getDeclaredMethod("methodB", int.class), 5);
    }

    @Test
    public void testNewInstance0() throws Exception {
        ClassD objectD = ReflectionHelper.newInstance(ClassD.class);
        assertNotNull(objectD);
    }

    @Test(expected = UncheckedException.class)
    public void testNewInstance1() throws Exception {
        ReflectionHelper.newInstance(ClassB.class);
    }

    @Test(expected = UncheckedException.class)
    public void testNewInstance2() throws Exception {
        ReflectionHelper.newInstance(ClassA.class);
    }

    @Test(expected = UncheckedException.class)
    public void testNewInstance3() throws Exception {
        ReflectionHelper.newInstance(ClassB.class);
    }

    @Test
    public void testLoadClassWithNullClassloader() throws Exception {
        Class<ReflectionHelper> a = ReflectionHelper.loadClass(ReflectionHelper.class.getName(),
                null);
        assertTrue(a.getName().equals(ReflectionHelper.class.getName()));
    }

    @Test
    public void testLoadClassWithClassloader() throws Exception {
        Class<ReflectionHelper> a = ReflectionHelper.loadClass(ReflectionHelper.class.getName(),
                ReflectionHelperTest.class.getClassLoader());
        assertTrue(a.getName().equals(ReflectionHelper.class.getName()));
    }

    @Test(expected = UncheckedException.class)
    public void testLoadClassWithUnknownClass() throws Exception {
        ReflectionHelper.loadClass("UnknownClass", null);
    }

    @Test(expected = UncheckedException.class)
    public void testGetFieldValue0() throws Exception {
        ClassD objectD = ReflectionHelper.newInstance(ClassD.class);
        Field field1 = ClassD.class.getDeclaredField(ClassD.PRIVATE_FIELD_NAME);
        ReflectionHelper.getFieldValue(objectD, field1);
    }

    @Test(expected = UncheckedException.class)
    public void testGetFieldValue1() throws Exception {
        ClassA objectA = ReflectionHelper.newInstance(ClassA.class);
        Field field1 = ClassD.class.getDeclaredField(ClassD.PRIVATE_FIELD_NAME);
        ReflectionHelper.getFieldValue(objectA, field1);
    }

    @Test
    public void testGetFieldValue2() throws Exception {
        ClassD objectD = ReflectionHelper.newInstance(ClassD.class);
        Field field1 = ClassD.class.getDeclaredField(ClassD.PRIVATE_FIELD_NAME);
        field1.setAccessible(true);
        assertEquals(ReflectionHelper.getPrivateFieldValue(ClassD.class, null, ClassD.STATIC_PRIVATE_FIELD_NAME),
                ReflectionHelper.getFieldValue(objectD, field1));
    }

    @Test(expected = UncheckedException.class)
    public void testSetFieldValue0() throws Exception {
        ClassD objectD = ReflectionHelper.newInstance(ClassD.class);
        Field field1 = ClassD.class.getDeclaredField(ClassD.PRIVATE_FIELD_NAME);
        ReflectionHelper.setFieldValue(objectD, field1, 5);
    }

    @Test(expected = UncheckedException.class)
    public void testSetFieldValue1() throws Exception {
        ClassA objectA = ReflectionHelper.newInstance(ClassA.class);
        Field field1 = ClassD.class.getDeclaredField(ClassD.PRIVATE_FIELD_NAME);
        ReflectionHelper.setFieldValue(objectA, field1, 5);
    }

    @Test(expected = UncheckedException.class)
    public void testSetFieldValue2() throws Exception {
        ClassD objectD = ReflectionHelper.newInstance(ClassD.class);
        Field field1 = ClassD.class.getDeclaredField(ClassD.PRIVATE_FIELD_NAME);
        field1.setAccessible(true);
        ReflectionHelper.setFieldValue(objectD, field1, "5");
    }

    @Test
    public void testSetFieldValue3() throws Exception {
        ClassD objectD = ReflectionHelper.newInstance(ClassD.class);
        Field field1 = ClassD.class.getDeclaredField(ClassD.PRIVATE_FIELD_NAME);
        field1.setAccessible(true);
        ReflectionHelper.setFieldValue(objectD, field1, 5);
    }

    @Test
    public void testIsOrInheritedFromWithAClassAndItsCanonicalName() {
        assertTrue(ReflectionHelper.isOrInheritedFrom(ClassA.class, this.getClass()
                .getCanonicalName() + ".ClassA"));
    }

    @Test
    public void testIsOrInheritedFromWithAClassAndACanonicalNameWhichIsNotOneOfItsSuperclassName() {
        assertTrue(!ReflectionHelper.isOrInheritedFrom(ClassA.class, this.getClass()
                .getCanonicalName() + ".ClassX"));
    }

    @Test
    public void testIsOrInheritedFromWithAClassAndACanonicalNameOfOneOfItsSuperclass() {
        assertTrue(ReflectionHelper.isOrInheritedFrom(ClassC.class, this.getClass()
                .getCanonicalName() + ".ClassA"));
    }

    @Test
    public void testIsOrInheritedFromWithAInterfaceAndACanonicalNameOfOneOfItsSuperclass() {
        InterfaceE interfaceE = new ClassC();
        assertTrue(ReflectionHelper.isOrInheritedFrom(interfaceE.getClass(), this.getClass()
                .getCanonicalName() + ".ClassA"));
    }

    @Test
    public void testSetAndGetPrivateFieldValue_0() throws Exception {
        ClassD objectD = ReflectionHelper.newInstance(ClassD.class);
        assertNotNull(objectD);
        final int expectedValue = 100;
        ReflectionHelper.setPrivateField(objectD, ClassD.PRIVATE_FIELD_NAME, expectedValue);
        final int value = (Integer) ReflectionHelper.getPrivateFieldValue(ClassD.class, objectD,
                ClassD.PRIVATE_FIELD_NAME);
        assertEquals(expectedValue, value);
    }

    /**
     * Unite test about {@link ReflectionHelper#setFieldValue(Class, Object, String, Object, boolean)} and
     * {@link ReflectionHelper#getFieldValue(Class, Object, String, boolean)}
     */
    @Test
    public void setField() throws Exception {
        final ClassD objectD = ReflectionHelper.newInstance(ClassD.class);
        assertNotNull(objectD);

        // setField on static final public field
        final int expectedStaticFinalPublicValue = 10;
        ReflectionHelper.setFieldValue(ClassD.class, null, ClassD.STATIC_FINAL_PUBLIC_FIELD_NAME,
                expectedStaticFinalPublicValue, true);
        assertEquals(expectedStaticFinalPublicValue,
                ReflectionHelper.getFieldValue(ClassD.class, null, ClassD.STATIC_FINAL_PUBLIC_FIELD_NAME, false));

        // setField on static final protected field
        final int expectedStaticFinalProtectedValue = 20;
        ReflectionHelper.setFieldValue(ClassD.class, null, ClassD.STATIC_FINAL_PROTECTED_FIELD_NAME,
                expectedStaticFinalProtectedValue, true);
        assertEquals(expectedStaticFinalProtectedValue,
                ReflectionHelper.getFieldValue(ClassD.class, null, ClassD.STATIC_FINAL_PROTECTED_FIELD_NAME, false));

        // setField on static final private field
        final int expectedStaticFinalPrivateValue = 30;
        ReflectionHelper.setFieldValue(ClassD.class, null, ClassD.STATIC_FINAL_PRIVATE_FIELD_NAME,
                expectedStaticFinalPrivateValue, true);
        assertEquals(expectedStaticFinalPrivateValue,
                ReflectionHelper.getFieldValue(ClassD.class, null, ClassD.STATIC_FINAL_PRIVATE_FIELD_NAME, false));

        // setField on static public field
        final int expectedStaticPublicValue = 100;
        ReflectionHelper
                .setFieldValue(ClassD.class, null, ClassD.STATIC_PUBLIC_FIELD_NAME, expectedStaticPublicValue, false);
        assertEquals(expectedStaticPublicValue,
                ReflectionHelper.getFieldValue(ClassD.class, null, ClassD.STATIC_PUBLIC_FIELD_NAME, false));

        // setField on static protected field
        final int expectedStaticProtectedValue = 200;
        ReflectionHelper.setFieldValue(ClassD.class, null, ClassD.STATIC_PROTECTED_FIELD_NAME, expectedStaticProtectedValue,
                false);
        assertEquals(expectedStaticProtectedValue,
                ReflectionHelper.getFieldValue(ClassD.class, null, ClassD.STATIC_PROTECTED_FIELD_NAME, false));

        // setField on static private field
        final int expectedStaticPrivateValue = 300;
        ReflectionHelper.setFieldValue(ClassD.class, null, ClassD.STATIC_PRIVATE_FIELD_NAME, expectedStaticPrivateValue,
                false);
        assertEquals(expectedStaticPrivateValue,
                ReflectionHelper.getFieldValue(ClassD.class, null, ClassD.STATIC_PRIVATE_FIELD_NAME, false));

        // setField on final public field
        final int expectedFinalPublicValue = 1000;
        ReflectionHelper
                .setFieldValue(ClassD.class, objectD, ClassD.FINAL_PUBLIC_FIELD_NAME, expectedFinalPublicValue, true);
        assertEquals(expectedFinalPublicValue,
                ReflectionHelper.getFieldValue(ClassD.class, objectD, ClassD.FINAL_PUBLIC_FIELD_NAME, false));

        // setField on final protected field
        final int expectedFinalProtectedValue = 2000;
        ReflectionHelper.setFieldValue(ClassD.class, objectD, ClassD.FINAL_PROTECTED_FIELD_NAME,
                expectedFinalProtectedValue, true);
        assertEquals(expectedFinalProtectedValue,
                ReflectionHelper.getFieldValue(ClassD.class, objectD, ClassD.FINAL_PROTECTED_FIELD_NAME, false));

        // setField on final private field
        final int expectedFinalPrivateValue = 3000;
        ReflectionHelper.setFieldValue(ClassD.class, objectD, ClassD.FINAL_PRIVATE_FIELD_NAME, expectedFinalPrivateValue,
                true);
        assertEquals(expectedFinalPrivateValue,
                ReflectionHelper.getFieldValue(ClassD.class, objectD, ClassD.FINAL_PRIVATE_FIELD_NAME, false));

        // setField on public field
        final int expectedPublicValue = 10000;
        ReflectionHelper.setFieldValue(ClassD.class, objectD, ClassD.PUBLIC_FIELD_NAME, expectedPublicValue, false);
        assertEquals(expectedPublicValue,
                ReflectionHelper.getFieldValue(ClassD.class, objectD, ClassD.PUBLIC_FIELD_NAME, false));

        // setField on protected field
        final int expectedProtectedValue = 20000;
        ReflectionHelper.setFieldValue(ClassD.class, objectD, ClassD.PROTECTED_FIELD_NAME, expectedProtectedValue, false);
        assertEquals(expectedProtectedValue,
                ReflectionHelper.getFieldValue(ClassD.class, objectD, ClassD.PROTECTED_FIELD_NAME, false));

        // setField on private field
        final int expectedPrivateValue = 30000;
        ReflectionHelper.setFieldValue(ClassD.class, objectD, ClassD.PRIVATE_FIELD_NAME, expectedPrivateValue, false);
        assertEquals(expectedPrivateValue,
                ReflectionHelper.getFieldValue(ClassD.class, objectD, ClassD.PRIVATE_FIELD_NAME, false));
        
        // setField on a final static attribute that was previously get without removing the modifier final
        ReflectionHelper.getFieldValue(ClassD.class, null, ClassD.STATIC_FINAL_PUBLIC_RO_1_FIELD_NAME, false);
        try {
        ReflectionHelper.setFieldValue(ClassD.class, null, ClassD.STATIC_FINAL_PUBLIC_RO_1_FIELD_NAME,
                expectedStaticFinalPublicValue, true);
            fail("An exception should occur about illegal access");
        } catch (final UncheckedException e) {
            assertTrue(e.getCause() instanceof IllegalAccessException);
        }
        // setField on a final static attribute that was previously get with removing the modifier final
        ReflectionHelper.getFieldValue(ClassD.class, null, ClassD.STATIC_FINAL_PUBLIC_RO_2_FIELD_NAME, true);
        ReflectionHelper.setFieldValue(ClassD.class, null, ClassD.STATIC_FINAL_PUBLIC_RO_2_FIELD_NAME,
                expectedStaticFinalPublicValue, true);
    }
}
